#![feature(random)]

use clap::Parser;
use sha2::{Digest, Sha512};
use std::io::{ErrorKind, prelude::*};
use std::net::{TcpListener, TcpStream};
use std::random;
use std::sync::{Arc, Mutex};
use std::thread::{self, JoinHandle, sleep};

#[derive(Parser, Debug)]
#[command(version, about, long_about = None)]
struct Args {
    /// Port to bind on.
    #[arg(short, long, default_value_t = 7867)]
    port: u16,
    /// Password. Will be stored as a SHA512 hashed string.
    /// Will not require a password if none provided.
    #[arg(short = 'P', long, default_value = None)]
    password: Option<String>
}

#[derive(Clone, PartialEq, Eq)]
enum UserStatus {
    Online,
    Disconnected,
    Idle,
    Custom(String),
}

impl UserStatus {
    fn get_status_string(&self) -> String {
        match self {
            UserStatus::Online => "\x1b[0m\x1b[32monline\x1b[0m".to_string(),
            UserStatus::Idle => "\x1b[0m\x1b[33midle\x1b[0m".to_string(),
            UserStatus::Disconnected => "\x1b[0m\x1b[31mdisconnected\x1b[0m".to_string(),
            UserStatus::Custom(val) => format!("\x1b[0m\x1b[34m{}\x1b[0m", val),
        }
    }
}

#[derive(Clone)]
struct User {
    name: String,
    status: UserStatus,
    token: String,
    no_show_msg: bool,
}

#[derive(Clone)]
struct Message {
    sender: User,
    contents: String,
    timestamp: time::OffsetDateTime,
    me_message: bool,
}

type Channel = Arc<Mutex<UnwrappedChannel>>;

#[derive(Clone)]
struct UnwrappedChannel {
    name: String,
    messages: Vec<Message>,
    online_users: Vec<User>,
}

fn display_msg(stream: &mut TcpStream, msg: &Message) -> bool {

    let time_format = time::format_description::well_known::Rfc2822;
    if !msg.me_message {
        if stream
            .write_fmt(format_args!(
                "\x1b[0m\x1b[34m[{}]\x1b[0m \x1b[1m{}\x1b[0m \x1b[31m⇛\x1b[0m \x1b[35m{}\x1b[0m\n",
                msg.timestamp.format(&time_format).expect("the time is evil"),
                msg.sender.name,
                msg.contents
            ))
            .is_err() {
                return false;
            }
    } else if stream
            .write_fmt(format_args!(
                "\x1b[0m\x1b[34m[{}]\x1b[0m \x1b[1;3m{}\x1b[0m \x1b[3m{}\x1b[0m\n",
                msg.timestamp.format(&time_format).expect("the time is evil"),
                msg.sender.name,
                msg.contents
            ))
            .is_err() {
                return false;
            }
    true
}

fn handle_client(mut stream: TcpStream, channel: Channel, password: Option<Vec<u8>>) {
    stream
        .write_all(b"\x1b[33mEnter your chosen nickname(up to 30 characters):\x1b[0m\x1b[4m\n")
        .unwrap();
    let mut buf = [0u8; 30];
    let mut i = 0usize;
    let mut buf1 = [0u8];
    'bufreadloop: while buf1[0] != b'\n' || i == 0 {
        stream.read_exact(&mut buf1).unwrap();
        if buf1[0] == b'\n' && i == 0 {
            stream.write_all(
                b"\x1b[0m\x1b[31;1mEnter at least one character.\n\x1b[0m\x1b[33mEnter your chosen nickname(up to 30 characters):\x1b[0m\x1b[4m\n"
            ).unwrap();
            buf = [0u8; 30];
            buf1[0] = b'\0';
            continue;
        } else if buf1[0] == b'\n' && i >= 30 {
            stream.write_all(
                b"\x1b[0m\x1b[31;1mEnter less than 30 characters.\n\x1b[0m\x1b[33mEnter your chosen nickname(up to 30 characters):\x1b[0m\x1b[4m\n"
            ).unwrap();
            buf = [0u8; 30];
            buf1[0] = b'\0';
            i = 0;
            continue;
        } else if buf1[0] == b'\n' {
            let new_name = String::from_utf8_lossy(&(buf[..i])).to_string();
            if new_name.to_lowercase() == "netbot" {
                stream.write_all(
                    b"\x1b[0m\x1b[31;1mNetbot: Hey! Rude.\n \
                    \x1b[0m\x1b[33mEnter your chosen nickname(up to 30 characters):\x1b[0m\x1b[4m\n"
                ).unwrap();
                buf = [0u8; 30];
                buf1[0] = b'\0';
                i = 0;
                continue 'bufreadloop;
            }
            let users;
            {
                users = channel.lock().unwrap().online_users.clone();
            }
            for user in users {
                if user.name == new_name {
                    if user.status == UserStatus::Disconnected {
                        stream
                            .write_all(b"\x1b[0m\x1b[33mEnter your login token(up to 50 characters):\x1b[0m\x1b[4m\n")
                            .unwrap();
                        let mut buf = [0u8; 50];
                        let mut i = 0usize;
                        let mut buf1 = [0u8];
                        'bufreadloop2: while buf1[0] != b'\n' || i == 0 {
                            stream.read_exact(&mut buf1).unwrap();
                            if buf1[0] == b'\n' && i >= 50 {
                                stream.write_all(
                                    b"\x1b[0m\x1b[31;1mEnter less than 50 characters.\n \
                                    \x1b[0m\x1b[33mEnter your login token(up to 50 characters):\x1b[0m\x1b[4m\n"
                                ).unwrap();
                                buf = [0u8; 50];
                                buf1[0] = b'\0';
                                i = 0;
                                continue;
                            } else if buf1[0] == b'\n' {
                                if user.token != String::from_utf8_lossy(&(buf[..i])) {
                                    break 'bufreadloop2;
                                }
                                break 'bufreadloop;
                            }
                            if i < 50 {
                                buf[i] = buf1[0];
                                i += 1;
                            }
                        }
                    }
                    stream.write_all(
                        b"\x1b[0m\x1b[31;1mSorry, but that nickname is taken.\n \
                        \x1b[0m\x1b[33mEnter your chosen nickname(up to 30 characters):\x1b[0m\x1b[4m\n"
                    ).unwrap();
                    buf = [0u8; 30];
                    buf1[0] = b'\0';
                    i = 0;
                    continue 'bufreadloop;
                }
            }
            break;
        }
        if i < 30 {
            buf[i] = buf1[0];
            i += 1;
        }
    }

    if password.is_some() {
        let password = password.unwrap();

        stream
            .write_all(b"\x1b[0m\x1b[33mThis server requires a password. Enter the password:\x1b[0m\x1b[4m\n")
            .unwrap();

        let mut buf = vec![];
        let mut buf1 = [0u8];
        'bufreadloop3: while buf1[0] != b'\n' || i == 0 {
            stream.read_exact(&mut buf1).unwrap();
            if buf1[0] == b'\n' {
                let mut sha512 = Sha512::new();
                sha512.update(buf);

                if password != sha512.finalize().to_vec() {
                    stream
                        .write_all(b"\x1b[0m\x1b[31mIncorrect.\x1b[0m\n")
                        .unwrap();
                    
                    return;
                }
                break 'bufreadloop3;
            }
            buf.push(buf1[0]);
        }
    }

    {
        let lock = channel.lock().unwrap();
        for msg in &lock.messages {
            if !display_msg(&mut stream, msg) {
                break;
            }
        }
    }

    stream
        .write_fmt(format_args!(
            "\x1b[0m\x1b[32;1mWelcome, {}! Joining server \"{}\".\x1b[0m\n",
            String::from_utf8_lossy(&(buf[..i])),
            channel.lock().unwrap().name
        ))
        .unwrap();

    let mut user = User {
        name: String::from_utf8_lossy(&(buf[..i])).to_string(),
        status: UserStatus::Online,
        token: format!("{:x}", random::random::<u128>()),
        no_show_msg: false
    };

    stream
        .write_fmt(format_args!(
            "\x1b[0m\x1b[32;1mYour token to rejoin after quitting is {}.\x1b[0m\n",
            user.token
        ))
        .unwrap();

    let netbot = User {
        name: "Netbot".to_owned(),
        status: UserStatus::Online,
        token: "".to_owned(),
        no_show_msg: false
    };

    let mut broke = false;
    for usr in &mut channel.lock().unwrap().online_users {
        if usr.name == user.name {
            usr.status = user.status.clone();
            usr.token = user.token.clone();
            user.no_show_msg = usr.no_show_msg;
            broke = true;
            break;
        }
    }

    if !broke {
        channel.lock().unwrap().online_users.push(user.clone());
    }

    let (recieve_terminate_send, recieve_terminate_recv) = std::sync::mpsc::channel::<()>();
    let recieve_thread: JoinHandle<_>;

    {
        let channel = channel.clone();
        let mut stream = stream.try_clone().unwrap();

        recieve_thread = thread::spawn(move || {
            let mut current_message_count = channel.lock().unwrap().messages.len();
            loop {
                if recieve_terminate_recv.try_recv().is_ok() {
                    let _ = stream.shutdown(std::net::Shutdown::Both);
                    panic!("recieve thread exiting");
                }

                #[warn(clippy::panic, clippy::unwrap_used)]
                {
                    #[allow(clippy::unwrap_used)]
                    let lock = channel.lock().unwrap();
                    let new_message_count = lock.messages.len();
                    if current_message_count != new_message_count {
                        let new_messages = &lock.messages[current_message_count..];
                        for msg in new_messages {
                            if !display_msg(&mut stream, msg) {
                                break;
                            }
                        }

                        current_message_count = new_message_count;
                    }
                }

                // so that not all the threads will sync at once
                let randomval = random::random::<u64>() / (18446744073709551615u64 / 20);
                sleep(std::time::Duration::from_millis(240 + randomval));
            }
        });
    }

    channel.lock().unwrap().messages.push(Message {
        sender: netbot.clone(),
        contents: format!("User {} has joined!", user.name),
        timestamp: time::OffsetDateTime::now_local().unwrap(),
        me_message: false,
    });

    let mut buf = Vec::<u8>::new();
    let mut buf1 = [0u8];

    loop {
        if let Err(err) = stream.read_exact(&mut buf1) {
            if err.kind() == ErrorKind::UnexpectedEof {
                user.status = UserStatus::Disconnected;
                for usr in &mut channel.lock().unwrap().online_users {
                    if usr.name == user.name {
                        usr.status = user.status.clone();
                        break;
                    }
                }

                channel.lock().unwrap().messages.push(Message {
                    sender: netbot.clone(),
                    contents: format!("User {} has left.", user.name),
                    timestamp: time::OffsetDateTime::now_local().unwrap(),
                    me_message: false,
                });

                recieve_terminate_send.send(()).unwrap();
                let _ = recieve_thread.join();
                return;
            }
        }

        if buf1[0] == b'\x08' {
            buf.pop();
        }

        if buf1[0] == b'\n' && !buf.is_empty() {
            let contents = String::from_utf8_lossy(&buf).to_string();
            if !contents.starts_with("/") {
                channel.lock().unwrap().messages.push(Message {
                    sender: user.clone(),
                    contents,
                    timestamp: time::OffsetDateTime::now_local().unwrap(),
                    me_message: false,
                });
            } else if contents.split(" ").collect::<Vec<&str>>()[0] == "/help"
                || contents.split(" ").collect::<Vec<&str>>()[0] == "/?"
            {
                stream
                        .write_all(
                            b"\x1b[0m\x1b[32mValid commands:\n \
                                /help, /?: Print this help.\n \
                                /status [online|idle|disconnected|(any value)]: Set your status.\n \
                                /user [nick]: Get information about a user.\n \
                                /server [new name]: Get or set the server name.\n \
                                /users: List users.\n \
                                /me x: Say that you are doing x(for example, \"/me purrs\" as netbot would result in the output of \"netbot purrs\").\n \
                                /clear: Clears showing all messages(including rejoining) until /show is ran.\n \
                                /show: Output all messages ever.\x1b[0m\n",
                        )
                        .unwrap();
            } else if contents.split(" ").collect::<Vec<&str>>()[0] == "/me" {
                if contents.split(" ").collect::<Vec<&str>>().len() < 2
                    || contents.split(" ").collect::<Vec<&str>>()[1].is_empty()
                {
                    stream
                            .write_fmt(format_args!(
                                "\x1b[0m\x1b[31mYou need to provide at least one argument to /me!\n\x1b[0m",
                            ))
                            .unwrap();
                } else {
                    channel.lock().unwrap().messages.push(Message {
                        sender: user.clone(),
                        contents: contents.replace("/me ", ""),
                        timestamp: time::OffsetDateTime::now_local().unwrap(),
                        me_message: true,
                    });
                }
            } else if contents.split(" ").collect::<Vec<&str>>()[0] == "/clear" {
                stream.write_all(b"\x1b[2J").unwrap();
                user.no_show_msg = true;
                for usr in &mut channel.lock().unwrap().online_users {
                    if usr.name == user.name {
                        usr.no_show_msg = user.no_show_msg;
                    }
                }
            } else if contents.split(" ").collect::<Vec<&str>>()[0] == "/show" {
                {
                    let lock = channel.lock().unwrap();
                    for msg in &lock.messages {
                        if !display_msg(&mut stream, msg) {
                            break;
                        }
                    }
                }
                user.no_show_msg = false;
                for usr in &mut channel.lock().unwrap().online_users {
                    if usr.name == user.name {
                        usr.no_show_msg = user.no_show_msg;
                    }
                }

            } else if contents.split(" ").collect::<Vec<&str>>()[0] == "/status"
                || contents == "/user"
            {
                let contents = contents.split(" ").collect::<Vec<&str>>();
                if contents.len() == 1 {
                    stream
                        .write_fmt(format_args!(
                            "\x1b[0m\x1b[32mYou are currently marked as {}\n\x1b[0m",
                            user.clone().status.get_status_string()
                        ))
                        .unwrap();
                } else {
                    match contents[1] {
                        "online" => {
                            user.status = UserStatus::Online;
                        }
                        "idle" => {
                            user.status = UserStatus::Idle;
                        }
                        "disconnected" => {
                            user.status = UserStatus::Disconnected;
                        }
                        other => {
                            user.status = UserStatus::Custom(other.to_owned());
                        }
                    }
                    for usr in &mut channel.lock().unwrap().online_users {
                        if usr.name == user.name {
                            usr.status = user.status.clone();
                        }
                    }
                    stream
                        .write_fmt(format_args!(
                            "\x1b[0m\x1b[32mYou are now marked as {}\n\x1b[0m",
                            user.status.get_status_string()
                        ))
                        .unwrap();
                }
            } else if contents.split(" ").collect::<Vec<&str>>()[0] == "/server" {
                let contents = contents.split(" ").collect::<Vec<&str>>();
                if contents.len() == 1 {
                    stream
                        .write_fmt(format_args!(
                            "\x1b[0m\x1b[32mCurrent server name: {}.\n\x1b[0m",
                            channel.lock().unwrap().name
                        ))
                        .unwrap();
                } else {
                    channel.lock().unwrap().name = contents[1].to_owned();
                    stream
                        .write_fmt(format_args!(
                            "\x1b[0m\x1b[32mNew server name: {}.\n\x1b[0m",
                            channel.lock().unwrap().name
                        ))
                        .unwrap();
                }
            } else if contents.split(" ").collect::<Vec<&str>>()[0] == "/user" {
                let contents = contents.split(" ").collect::<Vec<&str>>();
                let mut found_user = false;
                for usr in &mut channel.lock().unwrap().online_users {
                    if usr.name == contents[1] {
                        stream
                            .write_fmt(format_args!(
                                "\x1b[0m\x1b[32mUser {} is currently {}.\n\x1b[0m",
                                usr.name,
                                usr.status.get_status_string()
                            ))
                            .unwrap();
                        found_user = true;
                        break;
                    }
                }
                if !found_user {
                    stream
                        .write_fmt(format_args!(
                            "\x1b[0m\x1b[32mUnknown user \"{}\".\n\x1b[0m",
                            contents[1]
                        ))
                        .unwrap();
                }
            } else if contents.starts_with("/users") {
                stream.write_all(b"\x1b[0m\x1b[32mUsers:\x1b[0m").unwrap();
                for usr in &channel.lock().unwrap().online_users {
                    stream
                        .write_fmt(format_args!(
                            "\n{} (currently {})",
                            usr.name,
                            usr.status.get_status_string()
                        ))
                        .unwrap();
                }
                stream.write_all(b"\n\x1b[0m").unwrap();
            } else {
                stream
                    .write_all(b"\x1b[0m\x1b[31;1mThat's an invalid command!\n\x1b[0m")
                    .unwrap();
            }
            buf.clear();
        }
        if buf1[0] == b'\n' {
            continue;
        }
        buf.push(buf1[0]);
    }
}

fn main() -> std::io::Result<()> {
    let args = Args::parse();
    let mut password = None;
    if args.password.is_some() {
        let mut sha512 = sha2::Sha512::new();
        sha512.update(args.password.unwrap());
        password = Some(sha512.finalize().to_vec());
    }

    let listener = TcpListener::bind(format!("0.0.0.0:{}", args.port))?;
    let channel: Channel = Arc::new(Mutex::new(UnwrappedChannel {
        name: "Rawnetchat".to_string(),
        messages: vec![],
        online_users: vec![],
    }));

    for stream in listener.incoming() {
        let channel: Channel = channel.clone();
        let password = password.clone();
        thread::spawn(|| {
            handle_client(stream.unwrap(), channel, password);
        });
    }

    Ok(())
}
